"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OverlayWindow = void 0;
const events_1 = require("events");
const path_1 = require("path");
const throttle_debounce_1 = require("throttle-debounce");
const electron_1 = require("electron");
const lib = require('node-gyp-build')((0, path_1.join)(__dirname, '..'));
var EventType;
(function (EventType) {
    EventType[EventType["EVENT_ATTACH"] = 1] = "EVENT_ATTACH";
    EventType[EventType["EVENT_FOCUS"] = 2] = "EVENT_FOCUS";
    EventType[EventType["EVENT_BLUR"] = 3] = "EVENT_BLUR";
    EventType[EventType["EVENT_DETACH"] = 4] = "EVENT_DETACH";
    EventType[EventType["EVENT_FULLSCREEN"] = 5] = "EVENT_FULLSCREEN";
    EventType[EventType["EVENT_MOVERESIZE"] = 6] = "EVENT_MOVERESIZE";
})(EventType || (EventType = {}));
class OverlayWindow extends events_1.EventEmitter {
    constructor(overlayWindow, targetWindowTitle) {
        super();
        this.lastBounds = { x: 0, y: 0, width: 0, height: 0 };
        this.isFocused = false;
        if (OverlayWindow.isInitialized) {
            throw new Error('Library can be initialized only once.');
        }
        else {
            OverlayWindow.isInitialized = true;
        }
        this.electronWindow = overlayWindow;
        this.electronWindow.on('blur', () => {
            if (!this.isFocused && this.willBeFocused !== 'target') {
                this.electronWindow.hide();
            }
        });
        this.electronWindow.on('focus', () => {
            this.willBeFocused = undefined;
        });
        this.on('attach', (e) => {
            this.isFocused = true;
            this.electronWindow.setIgnoreMouseEvents(true);
            this.electronWindow.showInactive();
            if (process.platform === 'linux') {
                this.electronWindow.setSkipTaskbar(true);
            }
            this.electronWindow.setAlwaysOnTop(true, 'screen-saver');
            if (e.isFullscreen !== undefined) {
                this.electronWindow.setFullScreen(e.isFullscreen);
            }
            this.lastBounds = e;
            this.updateOverlayBounds();
        });
        this.on('fullscreen', (e) => {
            this.electronWindow.setFullScreen(e.isFullscreen);
        });
        this.on('detach', () => {
            this.isFocused = false;
            this.electronWindow.hide();
        });
        const dispatchMoveresize = (0, throttle_debounce_1.throttle)(34 /* 30fps */, this.updateOverlayBounds.bind(this));
        this.on('moveresize', (e) => {
            this.lastBounds = e;
            dispatchMoveresize();
        });
        this.on('blur', () => {
            this.isFocused = false;
            if (this.willBeFocused !== 'overlay' && !this.electronWindow.isFocused()) {
                this.electronWindow.hide();
            }
        });
        this.on('focus', () => {
            this.willBeFocused = undefined;
            this.isFocused = true;
            this.electronWindow.setIgnoreMouseEvents(true);
            if (!this.electronWindow.isVisible()) {
                this.electronWindow.showInactive();
                if (process.platform === 'linux') {
                    this.electronWindow.setSkipTaskbar(true);
                }
                this.electronWindow.setAlwaysOnTop(true, 'screen-saver');
            }
        });
        lib.start(this.electronWindow.getNativeWindowHandle(), targetWindowTitle, this.handler.bind(this));
    }
    updateOverlayBounds() {
        let lastBounds = this.lastBounds;
        if (lastBounds.width != 0 && lastBounds.height != 0) {
            if (process.platform === 'win32') {
                lastBounds = electron_1.screen.screenToDipRect(this.electronWindow, this.lastBounds);
            }
            this.electronWindow.setBounds(lastBounds);
            if (process.platform === 'win32') {
                // if moved to screen with different DPI, 2nd call to setBounds will correctly resize window
                // dipRect must be recalculated as well
                lastBounds = electron_1.screen.screenToDipRect(this.electronWindow, this.lastBounds);
                this.electronWindow.setBounds(lastBounds);
            }
        }
    }
    handler(e) {
        switch (e.type) {
            case EventType.EVENT_ATTACH:
                this.emit('attach', e);
                break;
            case EventType.EVENT_FOCUS:
                this.emit('focus', e);
                break;
            case EventType.EVENT_BLUR:
                this.emit('blur', e);
                break;
            case EventType.EVENT_DETACH:
                this.emit('detach', e);
                break;
            case EventType.EVENT_FULLSCREEN:
                this.emit('fullscreen', e);
                break;
            case EventType.EVENT_MOVERESIZE:
                this.emit('moveresize', e);
                break;
        }
    }
    activateOverlay() {
        this.willBeFocused = 'overlay';
        this.electronWindow.setIgnoreMouseEvents(false);
        this.electronWindow.focus();
    }
    focusTarget() {
        this.willBeFocused = 'target';
        lib.focusTarget();
    }
}
exports.OverlayWindow = OverlayWindow;
OverlayWindow.isInitialized = false;
OverlayWindow.WINDOW_OPTS = {
    fullscreenable: true,
    skipTaskbar: true,
    frame: false,
    show: false,
    transparent: true,
    // let Chromium to accept any size changes from OS
    resizable: true
};
//# sourceMappingURL=index.js.map